import ast
import logging

from django.contrib.contenttypes.models import ContentType
from django.db.models import Model
from rest_framework.decorators import action
from rest_framework.response import Response

from .utils import flatten_dict, serializer_data_diff, split_get
from .models import ADDITION, CHANGE, DELETION, OperationLogEntry
from .serializers import OperationLogEntrySerializer

logger = logging.getLogger(__name__)


class OperationLogMixin:
    """
    DRF 操作日志Mixin
    必须同时满足如下三个要求的操作才会记录
    1. 写操作(POST, PUT, PATCH, DELETE)
    2. 未被包含在`operationlog_action_exclude`中的action
    3. action 为 create, update, partial_update 或 destroy

    自定义action需要自己创建operation_log
    """

    CLEANED_SUBSTITUTE = "********************"
    sensitive_fields = {}
    operationlog_action_exclude = []
    operationlog_domain_field: str = None

    def _get_view_method(self, request):
        """Get view method."""
        if hasattr(self, "action"):
            return self.action or None
        return request.method.lower()

    @staticmethod
    def _get_user(request):
        """Get user."""
        user = request.user
        if user.is_anonymous:
            return None
        return user

    def _clean_data(self, data):
        """
        Clean a dictionary of data of potentially sensitive info before
        sending to the database.
        Function based on the "_clean_credentials" function of django
        (https://github.com/django/django/blob/stable/1.11.x/django/contrib/auth/__init__.py#L50)
        Fields defined by django are by default cleaned with this function
        You can define your own sensitive fields in your view by defining a set
        eg: sensitive_fields = {'field1', 'field2'}
        """
        if isinstance(data, bytes):
            data = data.decode(errors="replace")

        if isinstance(data, list):
            return [self._clean_data(d) for d in data]

        if isinstance(data, dict):
            SENSITIVE_FIELDS: set = {
                "api",
                "token",
                "key",
                "secret",
                "password",
                "signature",
            }

            data = dict(data)
            if self.sensitive_fields:
                SENSITIVE_FIELDS = SENSITIVE_FIELDS | {
                    field.lower() for field in self.sensitive_fields
                }

            for key, value in data.items():
                try:
                    value = ast.literal_eval(value)
                except (ValueError, SyntaxError):
                    pass
                if isinstance(value, (list, dict)):
                    data[key] = self._clean_data(value)
                if key.lower() in SENSITIVE_FIELDS:
                    data[key] = self.CLEANED_SUBSTITUTE
        return data

    def perform_create(self, serializer):
        super().perform_create(serializer)  # noqa

        request = self.request  # noqa
        if self.should_log(request):
            self._initial_log(
                request,
                serializer.instance,
                new_message=flatten_dict(serializer.validated_data),
            )

    def perform_update(self, serializer):
        request = self.request  # noqa
        if self.should_log(request):
            new_message = flatten_dict(serializer.validated_data)
            old_message = {}
            for k in new_message.keys():
                old_message[k] = split_get(serializer.instance, k)

            self._initial_log(
                request,
                serializer.instance,
                old_message=old_message,
                new_message=new_message,
                serializer=serializer,
            )

        super().perform_update(serializer)  # noqa

    def perform_destroy(self, instance):
        request = self.request  # noqa
        if self.should_log(request):
            self._initial_log(request, instance)

        super().perform_destroy(instance)  # noqa

    @action(
        detail=True,
        name="操作日志",
        serializer_class=OperationLogEntrySerializer,
    )
    def operationlogs(self, request, pk):
        """
        获取该资源操作日志接口
        :param request:
        :param pk: 主键
        :return:
        """
        queryset = OperationLogEntry.objects.select_related(
            "user", "content_type", "domain_content_type"
        ).filter(
            object_id=pk,
            content_type=ContentType.objects.get_for_model(self.queryset.model),
        )  # noqa
        queryset = self.filter_queryset(queryset)  # noqa
        page = self.paginate_queryset(queryset)  # noqa
        if page is not None:
            serializer = self.get_serializer(page, many=True)  # noqa
            return self.get_paginated_response(serializer.data)  # noqa

        serializer = self.get_serializer(queryset, many=True)  # noqa
        return Response(serializer.data)

    def should_log(self, request) -> bool:
        """
        是否记录操作日志, 可以覆盖此方法来自定义控制
        :param request:
        :return:
        """
        return (
            request.method in ("POST", "PUT", "PATCH", "DELETE")
            and self.action not in self.operationlog_action_exclude
        )

    def _get_action_name(self) -> str:
        """
        获取动作名称
        """
        if self.action == "create":
            return "新增"
        elif self.action in ("update", "partial_update"):
            return "编辑"
        elif self.action == "destroy":
            return "删除"
        else:
            return getattr(self, self.action).kwargs["name"]

    @staticmethod
    def _get_action_flag(request) -> int:
        if request.method == "POST":
            return ADDITION
        elif request.method in ("PUT", "PATCH"):
            return CHANGE
        elif request.method == "DELETE":
            return DELETION

    def _initial_log(
        self,
        request,
        instance,
        old_message=None,
        new_message=None,
        change_message=None,
        serializer=None,
    ) -> None:
        if change_message is None and old_message and new_message and serializer:
            change_message = serializer_data_diff(old_message, new_message, serializer)

        self.operation_log = OperationLogEntry(
            user=request.user,
            action=self.action,
            action_name=self._get_action_name(),
            action_flag=self._get_action_flag(request),
            content_type=ContentType.objects.get_for_model(instance),
            object_id=instance.pk,
            domain_content_type=ContentType.objects.get_for_model(instance),
            domain_object_id=instance.pk,
            change_message=change_message or {},
        )

        if self.operationlog_domain_field:
            attrs = self.operationlog_domain_field.split("__")
            obj = instance
            for attr in attrs:
                obj = getattr(obj, attr)

            if not isinstance(obj, Model):
                raise ValueError("'operationlog_domain_field' must refer to a model!")

            self.operation_log.domain_content_type = ContentType.objects.get_for_model(
                obj
            )
            self.operation_log.domain_object_id = obj.pk

    def finalize_response(self, request, response, *args, **kwargs):
        if (
            hasattr(self, "operation_log")
            and self.operation_log.action
            and self.operation_log.object_id
            and self.operation_log.content_type
            and not getattr(response, "exception", False)
        ):
            self.operation_log.save()

        return super().finalize_response(request, response, *args, **kwargs)



from rest_framework import serializers

from .models import OperationLogEntry


class OperationLogEntrySerializer(serializers.ModelSerializer):
    operator = serializers.CharField(default="", label="操作人")
    change_message = serializers.ListField(source="get_change_message", child=serializers.CharField(), label="操作内容")
    content_type_name = serializers.CharField(source="content_type.name", label="对象名称")
    domain_content_type_name = serializers.CharField(source="domain_content_type.name", default="",  label="域对象名称")

    class Meta:
        model = OperationLogEntry
        fields = "__all__"



from rest_framework.routers import SimpleRouter

from .views import OperationlogViewSet

router = SimpleRouter()

router.register("operationlogs", OperationlogViewSet, basename='operationlogs')
urlpatterns = router.urls



from django.core.validators import EMPTY_VALUES
from rest_framework.mixins import ListModelMixin, RetrieveModelMixin
from rest_framework.viewsets import GenericViewSet
from rest_framework.permissions import IsAdminUser

from .models import OperationLogEntry
from .serializers import (
    OperationLogEntrySerializer
)


class OperationlogViewSet(
    ListModelMixin,
    RetrieveModelMixin,
    GenericViewSet,
):
    queryset = OperationLogEntry.objects.select_related("user", "content_type", "domain_content_type")
    permission_classes = (IsAdminUser,)
    serializer_class = OperationLogEntrySerializer



import json

from django.conf import settings
from django.contrib.admin.utils import quote
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.urls import NoReverseMatch, reverse
from django.utils import timezone
from django.utils.text import get_text_list
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from .encoders import JSONEncoder

ADDITION = 1
CHANGE = 2
DELETION = 3

ACTION_FLAG_CHOICES = (
    (ADDITION, _("Addition")),
    (CHANGE, _("Change")),
    (DELETION, _("Deletion")),
)


class OperationLogEntryManager(models.Manager):
    use_in_migrations = True

    def log_action(
        self,
        user_id,
        content_type_id,
        object_id,
        object_repr,
        action_flag,
        change_message="",
    ):
        if isinstance(change_message, list):
            change_message = json.dumps(change_message)
        return self.model.objects.create(
            user_id=user_id,
            content_type_id=content_type_id,
            object_id=str(object_id),
            object_repr=object_repr[:200],
            action_flag=action_flag,
            change_message=change_message,
        )


class OperationLogEntry(models.Model):
    action_time = models.DateTimeField(
        _("操作时间"),
        default=timezone.now,
        db_index=True,
        editable=False,
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        models.CASCADE,
        verbose_name=_("用户"),
    )
    content_type = models.ForeignKey(
        ContentType,
        models.SET_NULL,
        verbose_name=_("操作对象"),
        related_name="+",
        blank=True,
        null=True,
    )
    object_id = models.TextField(_("对象ID"), blank=True, null=True)
    domain_content_type = models.ForeignKey(
        ContentType,
        models.SET_NULL,
        verbose_name=_("同范围操作对象"),
        related_name="+",
        blank=True,
        null=True,
    )
    domain_object_id = models.TextField(_("同范围对象ID"), blank=True, null=True)
    action = models.CharField(_("动作"), max_length=32)
    action_name = models.CharField(_("动作名称"), max_length=32)
    action_flag = models.PositiveSmallIntegerField(_("操作类型"), choices=ACTION_FLAG_CHOICES)
    change_message = models.JSONField(_("差异信息"), blank=True, default=dict, encoder=JSONEncoder)

    objects = OperationLogEntryManager()

    class Meta:
        verbose_name = verbose_name_plural = _("操作日志")
        db_table = "drf_operation_log"
        ordering = ["-action_time"]
        index_together = ("content_type", "object_id")

    def __repr__(self):
        return str(self.action_time)

    def __str__(self):
        if self.is_addition():
            return gettext("新增 “%(object)s”.") % {"object": self.object_repr}
        elif self.is_change():
            return gettext("修改 “%(object)s” — %(changes)s") % {
                "object": self.object_repr,
                "changes": self.get_change_message(),
            }
        elif self.is_deletion():
            return gettext("删除 “%(object)s.”") % {"object": self.object_repr}

        return gettext("LogEntry Object")

    @property
    def operator(self):
        if hasattr(self.user, "name") and self.user.name:
            return self.user.name

        return getattr(self.user, self.user.USERNAME_FIELD)

    def is_addition(self):
        return self.action_flag == ADDITION

    def is_change(self):
        return self.action_flag == CHANGE

    def is_deletion(self):
        return self.action_flag == DELETION

    def get_change_message(self):
        """
        If self.change_message is a JSON structure, interpret it as a change
        string, properly translated.
        """
        if self.is_addition():
            return [f"新增 {self.content_type.name}"]
        elif self.is_deletion():
            return [f"删除 {self.content_type.name}"]
        elif not self.change_message:
            return ["未更改"]

        ret = []
        for field_name, (old_value, new_value) in self.change_message.items():
            old_value = "" if old_value is None else old_value
            new_value = "" if new_value is None else new_value
            message = f"修改 {field_name}， 旧值“{old_value}”，新值“{new_value}”"
            ret.append(message)

        return ret

    def get_edited_object(self):
        """Return the edited object represented by this log entry."""
        return self.content_type.get_object_for_this_type(pk=self.object_id)

    def get_operation_url(self):
        """
        Return the admin URL to edit the object represented by this log entry.
        """
        if self.content_type and self.object_id:
            url_name = "operationlogs-list"
            try:
                url = reverse(url_name)
                return f"{url}?content_type={self.content_type.pk}&object_id={self.object_id}"
            except NoReverseMatch:
                pass
        return None



# Generated by Django 3.2.13 on 2022-04-28 11:16

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('drf_operation_log', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='operationlogentry',
            name='domain_content_type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='contenttypes.contenttype', verbose_name='同范围操作对象'),
        ),
        migrations.AddField(
            model_name='operationlogentry',
            name='domain_object_id',
            field=models.TextField(blank=True, null=True, verbose_name='同范围对象ID'),
        ),
        migrations.AlterField(
            model_name='operationlogentry',
            name='content_type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='contenttypes.contenttype', verbose_name='操作对象'),
        ),
    ]



from collections.abc import MutableMapping

from django.db.models import Manager
from rest_framework.fields import ChoiceField, BooleanField
from rest_framework.serializers import Serializer

attribute_sep = "<"
missing_value = object()


def split_get(o: object, concat_attr: str, sep: str = attribute_sep):
    for attr in concat_attr.split(sep):
        try:
            o = getattr(o, attr)
        except AttributeError:
            return missing_value

    return o


def _get_source_fields(fields):
    return {f.source: f for f in fields.values() if f.source != "*"}


def _flatten_dict_gen(d, parent_key, sep):
    for k, v in d.items():
        if getattr(v, "read_only", False):
            continue

        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            yield from flatten_dict(v, new_key, sep=sep).items()
        elif isinstance(v, Serializer):
            yield from flatten_dict(_get_source_fields(v.fields), new_key, sep=sep).items()
        else:
            yield new_key, v


def flatten_dict(d: MutableMapping, parent_key: str = "", sep: str = attribute_sep):
    return dict(_flatten_dict_gen(d, parent_key, sep))


def serializer_data_diff(old_data: dict, new_data: dict, serializer: Serializer) -> dict:
    """
    判断两个个序列化器校验后数据数据差异
    :return:
    """
    trans_dic = flatten_dict(_get_source_fields(serializer.fields))

    ret = {}

    for (k_old, v_old), (k_new, v_new) in zip(old_data.items(), new_data.items()):
        if k_old != k_new:
            raise ValueError("比较的数据key顺序错误")

        try:
            field = trans_dic[k_new]
            label = field.label or k_new
            if isinstance(field, ChoiceField):
                v_mapping = dict(field.choices)
                v_old = v_mapping.get(v_old) or v_old
                v_new = v_mapping.get(v_new) or v_new
            elif isinstance(v_old, Manager):
                v_old = list(v_old.all())
            elif isinstance(field, BooleanField):
                v_old = "是" if v_old else "否"
                v_new = "是" if v_new else "否"

            if v_old == v_new or v_old == missing_value:
                continue

            ret[label] = [v_old, v_new]
        except KeyError:
            pass
    return ret



"""
Helper classes for parsers.
"""
import datetime
import decimal
import json  # noqa
import uuid

from django.db.models import Model
from django.db.models.query import QuerySet
from django.utils import timezone
from django.utils.encoding import force_str
from django.utils.functional import Promise
from rest_framework.compat import coreapi


class JSONEncoder(json.JSONEncoder):
    """
    JSONEncoder subclass that knows how to encode date/time/timedelta,
    decimal types, generators and other basic python objects.
    """

    def default(self, obj):
        # For Date Time string spec, see ECMA 262
        # https://ecma-international.org/ecma-262/5.1/#sec-15.9.1.15
        if isinstance(obj, Promise):
            return force_str(obj)
        elif isinstance(obj, datetime.datetime):
            representation = obj.isoformat()
            if representation.endswith("+00:00"):
                representation = representation[:-6] + "Z"
            return representation
        elif isinstance(obj, datetime.date):
            return obj.isoformat()
        elif isinstance(obj, datetime.time):
            if timezone and timezone.is_aware(obj):
                raise ValueError("JSON can't represent timezone-aware times.")
            representation = obj.isoformat()
            return representation
        elif isinstance(obj, datetime.timedelta):
            return str(obj.total_seconds())
        elif isinstance(obj, decimal.Decimal):
            # Serializers will coerce decimals to strings by default.
            return float(obj)
        elif isinstance(obj, uuid.UUID):
            return str(obj)
        elif isinstance(obj, QuerySet):
            return tuple(obj)
        elif isinstance(obj, Model):
            return str(obj)
        elif isinstance(obj, bytes):
            # Best-effort for binary blobs. See #4187.
            return obj.decode()
        elif hasattr(obj, "tolist"):
            # Numpy arrays and array scalars.
            return obj.tolist()
        elif (coreapi is not None) and isinstance(obj, (coreapi.Document, coreapi.Error)):
            raise RuntimeError(
                "Cannot return a coreapi object from a JSON view. "
                "You should be using a schema renderer instead for this view."
            )
        elif hasattr(obj, "__getitem__"):
            cls = list if isinstance(obj, (list, tuple)) else dict
            try:
                return cls(obj)
            except Exception:
                pass
        elif hasattr(obj, "__iter__"):
            return tuple(item for item in obj)
        return super().default(obj)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['drf_operation_log', 'drf_operation_log.migrations']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.2,<4.0', 'djangorestframework>=3.13.1,<4.0.0']

setup_kwargs = {
    'name': 'drf-operation-log',
    'version': '0.3.22',
    'description': 'Operation log for drf serializers.',
    'long_description': None,
    'author': 'aiden_lu',
    'author_email': 'aiden_lu@wochacha.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/anyidea/drf-operation-log',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



# Generated by Django 3.2.12 on 2022-04-07 02:07

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import drf_operation_log.encoders
import drf_operation_log.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contenttypes", "__first__"),
    ]

    operations = [
        migrations.CreateModel(
            name='OperationLogEntry',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('action_time', models.DateTimeField(db_index=True, default=django.utils.timezone.now, editable=False, verbose_name='操作时间')),
                ('object_id', models.TextField(blank=True, null=True, verbose_name='对象ID')),
                ('action', models.CharField(max_length=32, verbose_name='动作')),
                ('action_name', models.CharField(max_length=32, verbose_name='动作名称')),
                ('action_flag', models.PositiveSmallIntegerField(choices=[(1, 'Addition'), (2, 'Change'), (3, 'Deletion')], verbose_name='操作类型')),
                ('change_message', models.JSONField(blank=True, default=dict, encoder=drf_operation_log.encoders.JSONEncoder, verbose_name='差异信息')),
                ('content_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='contenttypes.contenttype', verbose_name='操作对象')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='用户')),
            ],
            options={
                'verbose_name': '操作日志',
                'verbose_name_plural': '操作日志',
                'db_table': 'drf_operation_log',
                'ordering': ['-action_time'],
                'index_together': {('content_type', 'object_id')},
            },
            managers=[
                ('objects', drf_operation_log.models.OperationLogEntryManager()),
            ],
        ),
    ]
